"""
Management of Docker containers

.. versionadded:: 2017.7.0

:depends: docker_ Python module

.. note::
    Older releases of the Python bindings for Docker were called docker-py_ in
    PyPI. All releases of docker_, and releases of docker-py_ >= 1.6.0 are
    supported. These python bindings can easily be installed using
    :py:func:`pip.install <salt.modules.pip.install>`:

    .. code-block:: bash

        salt myminion pip.install docker

    To upgrade from docker-py_ to docker_, you must first uninstall docker-py_,
    and then install docker_:

    .. code-block:: bash

        salt myminion pip.uninstall docker-py
        salt myminion pip.install docker

.. _docker: https://pypi.python.org/pypi/docker
.. _docker-py: https://pypi.python.org/pypi/docker-py

These states were moved from the :mod:`docker <salt.states.docker>` state
module (formerly called **dockerng**) in the 2017.7.0 release. When running the
:py:func:`docker_container.running <salt.states.docker_container.running>`
state for the first time after upgrading to 2017.7.0, your container(s) may be
replaced. The changes may show diffs for certain parameters which say that the
old value was an empty string, and the new value is ``None``. This is due to
the fact that in prior releases Salt was passing empty strings for these values
when creating the container if they were undefined in the SLS file, where now
Salt simply does not pass any arguments not explicitly defined in the SLS file.
Subsequent runs of the state should not replace the container if the
configuration remains unchanged.


.. note::
    To pull from a Docker registry, authentication must be configured. See
    :ref:`here <docker-authentication>` for more information on how to
    configure access to docker registries in :ref:`Pillar <pillar>` data.
"""

import copy
import logging

import salt.utils.args
import salt.utils.data
import salt.utils.dockermod
from salt.exceptions import CommandExecutionError

# Enable proper logging
log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Define the module's virtual name
__virtualname__ = "docker_container"
__virtual_aliases__ = ("moby_container",)

__deprecated__ = (
    3009,
    "docker",
    "https://github.com/saltstack/saltext-docker",
)


def __virtual__():
    """
    Only load if the docker execution module is available
    """
    if "docker.version" in __salt__:
        return __virtualname__
    return (False, __salt__.missing_fun_string("docker.version"))


def _format_comments(ret, comments):
    """
    DRY code for joining comments together and conditionally adding a period at
    the end, and adding this comment string to the state return dict.
    """
    if isinstance(comments, str):
        ret["comment"] = comments
    else:
        ret["comment"] = ". ".join(comments)
        if len(comments) > 1:
            ret["comment"] += "."
    return ret


def _check_diff(changes):
    """
    Check the diff for signs of incorrect argument handling in previous
    releases, as discovered here:

    https://github.com/saltstack/salt/pull/39996#issuecomment-288025200
    """
    for conf_dict in changes:
        if conf_dict == "Networks":
            continue
        for item in changes[conf_dict]:
            if changes[conf_dict][item]["new"] is None:
                old = changes[conf_dict][item]["old"]
                if old == "":
                    return True
                else:
                    try:
                        if all(x == "" for x in old):
                            return True
                    except TypeError:
                        # Old value is not an iterable type
                        pass
    return False


def _parse_networks(networks):
    """
    Common logic for parsing the networks
    """
    networks = salt.utils.args.split_input(networks or [])
    if not networks:
        networks = {}
    else:
        # We don't want to recurse the repack, as the values of the kwargs
        # being passed when connecting to the network will not be dictlists.
        networks = salt.utils.data.repack_dictlist(networks)
        if not networks:
            raise CommandExecutionError(
                "Invalid network configuration (see documentation)"
            )
        for net_name, net_conf in networks.items():
            if net_conf is None:
                networks[net_name] = {}
            else:
                networks[net_name] = salt.utils.data.repack_dictlist(net_conf)
                if not networks[net_name]:
                    raise CommandExecutionError(
                        "Invalid configuration for network '{}' "
                        "(see documentation)".format(net_name)
                    )
                for key in ("links", "aliases"):
                    try:
                        networks[net_name][key] = salt.utils.args.split_input(
                            networks[net_name][key]
                        )
                    except KeyError:
                        continue

        # Iterate over the networks again now, looking for
        # incorrectly-formatted arguments
        errors = []
        for net_name, net_conf in networks.items():
            if net_conf is not None:
                for key, val in net_conf.items():
                    if val is None:
                        errors.append(
                            "Config option '{}' for network '{}' is "
                            "missing a value".format(key, net_name)
                        )
        if errors:
            raise CommandExecutionError("Invalid network configuration", info=errors)

    if networks:
        try:
            all_networks = [
                x["Name"] for x in __salt__["docker.networks"]() if "Name" in x
            ]
        except CommandExecutionError as exc:
            raise CommandExecutionError(
                f"Failed to get list of existing networks: {exc}."
            )
        else:
            missing_networks = [x for x in sorted(networks) if x not in all_networks]
            if missing_networks:
                raise CommandExecutionError(
                    "The following networks are not present: {}".format(
                        ", ".join(missing_networks)
                    )
                )

    return networks


def _resolve_image(ret, image, client_timeout):
    """
    Resolve the image ID and pull the image if necessary
    """
    image_id = __salt__["docker.resolve_image_id"](image)

    if image_id is False:
        if not __opts__["test"]:
            # Image not pulled locally, so try pulling it
            try:
                pull_result = __salt__["docker.pull"](
                    image,
                    client_timeout=client_timeout,
                )
            except Exception as exc:  # pylint: disable=broad-except
                raise CommandExecutionError(f"Failed to pull {image}: {exc}")
            else:
                ret["changes"]["image"] = pull_result
                # Try resolving again now that we've pulled
                image_id = __salt__["docker.resolve_image_id"](image)
                if image_id is False:
                    # Shouldn't happen unless the pull failed
                    raise CommandExecutionError(
                        "Image '{}' not present despite a docker pull "
                        "raising no errors".format(image)
                    )
    return image_id


def running(
    name,
    image=None,
    skip_translate=None,
    ignore_collisions=False,
    validate_ip_addrs=True,
    force=False,
    watch_action="force",
    start=True,
    shutdown_timeout=None,
    client_timeout=salt.utils.dockermod.CLIENT_TIMEOUT,
    networks=None,
    **kwargs,
):
    """
    Ensure that a container with a specific configuration is present and
    running

    name
        Name of the container

    image
        Image to use for the container

        .. note::
            This state will pull the image if it is not present. However, if
            the image needs to be built from a Dockerfile or loaded from a
            saved image, or if you would like to use requisites to trigger a
            replacement of the container when the image is updated, then the
            :py:func:`docker_image.present
            <salt.states.dockermod.image_present>` state should be used to
            manage the image.

        .. versionchanged:: 2018.3.0
            If no tag is specified in the image name, and nothing matching the
            specified image is pulled on the minion, the ``docker pull`` that
            retrieves the image will pull *all tags* for the image. A tag of
            ``latest`` is no longer implicit for the pull. For this reason, it
            is recommended to specify the image in ``repo:tag`` notation.

    .. _docker-container-running-skip-translate:

    skip_translate
        This function translates Salt CLI or SLS input into the format which
        docker-py_ expects. However, in the event that Salt's translation logic
        fails (due to potential changes in the Docker Remote API, or to bugs in
        the translation code), this argument can be used to exert granular
        control over which arguments are translated and which are not.

        Pass this argument as a comma-separated list (or Python list) of
        arguments, and translation for each passed argument name will be
        skipped. Alternatively, pass ``True`` and *all* translation will be
        skipped.

        Skipping tranlsation allows for arguments to be formatted directly in
        the format which docker-py_ expects. This allows for API changes and
        other issues to be more easily worked around. An example of using this
        option to skip translation would be:

        For example, imagine that there is an issue with processing the
        ``port_bindings`` argument, and the following configuration no longer
        works as expected:

        .. code-block:: yaml

            mycontainer:
              docker_container.running:
                - image: 7.3.1611
                - port_bindings:
                  - 10.2.9.10:8080:80

        By using ``skip_translate``, you can forego the input translation and
        configure the port binding in the format docker-py_ needs:

        .. code-block:: yaml

            mycontainer:
              docker_container.running:
                - image: 7.3.1611
                - skip_translate: port_bindings
                - port_bindings: {8080: [('10.2.9.10', 80)], '4193/udp': 9314}

        See the following links for more information:

        - `docker-py Low-level API`_
        - `Docker Engine API`_

    .. _docker-py: https://pypi.python.org/pypi/docker-py
    .. _`docker-py Low-level API`: http://docker-py.readthedocs.io/en/stable/api.html#docker.api.container.ContainerApiMixin.create_container
    .. _`Docker Engine API`: https://docs.docker.com/engine/api/v1.33/#operation/ContainerCreate

    ignore_collisions : False
        Since many of docker-py_'s arguments differ in name from their CLI
        counterparts (with which most Docker users are more familiar), Salt
        detects usage of these and aliases them to the docker-py_ version of
        that argument so that both CLI and API versions of a given argument are
        supported. However, if both the alias and the docker-py_ version of the
        same argument (e.g. ``env`` and ``environment``) are used, an error
        will be raised. Set this argument to ``True`` to suppress these errors
        and keep the docker-py_ version of the argument.

    validate_ip_addrs : True
        For parameters which accept IP addresses as input, IP address
        validation will be performed. To disable, set this to ``False``

    force : False
        Set this parameter to ``True`` to force Salt to re-create the container
        irrespective of whether or not it is configured as desired.

    watch_action : force
        Control what type of action is taken when this state :ref:`watches
        <requisites-watch>` another state that has changes. The default action
        is ``force``, which runs the state with ``force`` set to ``True``,
        triggering a rebuild of the container.

        If any other value is passed, it will be assumed to be a kill signal.
        If the container matches the specified configuration, and is running,
        then the action will be to send that signal to the container. Kill
        signals can be either strings or numbers, and are defined in the
        **Standard Signals** section of the ``signal(7)`` manpage. Run ``man 7
        signal`` on a Linux host to browse this manpage. For example:

        .. code-block:: yaml

            mycontainer:
              docker_container.running:
                - image: busybox
                - watch_action: SIGHUP
                - watch:
                  - file: some_file

        .. note::

            If the container differs from the specified configuration, or is
            not running, then instead of sending a signal to the container, the
            container will be re-created/started and no signal will be sent.

    start : True
        Set to ``False`` to suppress starting of the container if it exists,
        matches the desired configuration, but is not running. This is useful
        for data-only containers, or for non-daemonized container processes,
        such as the Django ``migrate`` and ``collectstatic`` commands. In
        instances such as this, the container only needs to be started the
        first time.

    shutdown_timeout
        If the container needs to be replaced, the container will be stopped
        using :py:func:`docker.stop <salt.modules.dockermod.stop>`. If a
        ``shutdown_timout`` is not set, and the container was created using
        ``stop_timeout``, that timeout will be used. If neither of these values
        were set, then a timeout of 10 seconds will be used.

        .. versionchanged:: 2017.7.0
            This option was renamed from ``stop_timeout`` to
            ``shutdown_timeout`` to accommodate the ``stop_timeout`` container
            configuration setting.

    client_timeout : 60
        Timeout in seconds for the Docker client. This is not a timeout for
        this function, but for receiving a response from the API.

        .. note::
            This is only used if Salt needs to pull the requested image.

    .. _salt-states-docker-container-network-management:

    **NETWORK MANAGEMENT**

    .. versionadded:: 2018.3.0
    .. versionchanged:: 2019.2.0
        If the ``networks`` option is used, any networks (including the default
        ``bridge`` network) which are not specified will be disconnected.

    The ``networks`` argument can be used to ensure that a container is
    attached to one or more networks. Optionally, arguments can be passed to
    the networks. In the example below, ``net1`` is being configured with
    arguments, while ``net2`` and ``bridge`` are being configured *without*
    arguments:

    .. code-block:: yaml

        foo:
          docker_container.running:
            - image: myuser/myimage:foo
            - networks:
              - net1:
                - aliases:
                  - bar
                  - baz
                - ipv4_address: 10.0.20.50
              - net2
              - bridge
            - require:
              - docker_network: net1
              - docker_network: net2

    The supported arguments are the ones from the docker-py's
    `connect_container_to_network`_ function (other than ``container`` and
    ``net_id``).

    .. important::
        Unlike with the arguments described in the **CONTAINER CONFIGURATION
        PARAMETERS** section below, these network configuration parameters are
        not translated at all.  Consult the `connect_container_to_network`_
        documentation for the correct type/format of data to pass.

    .. _`connect_container_to_network`: https://docker-py.readthedocs.io/en/stable/api.html#docker.api.network.NetworkApiMixin.connect_container_to_network

    To start a container with no network connectivity (only possible in
    2019.2.0 and later) pass this option as an empty list. For example:

    .. code-block:: yaml

        foo:
          docker_container.running:
            - image: myuser/myimage:foo
            - networks: []


    **CONTAINER CONFIGURATION PARAMETERS**

    auto_remove (or *rm*) : False
        Enable auto-removal of the container on daemon side when the
        container’s process exits (analogous to running a docker container with
        ``--rm`` on the CLI).

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - auto_remove: True

    binds
        Files/directories to bind mount. Each bind mount should be passed in
        one of the following formats:

        - ``<host_path>:<container_path>`` - ``host_path`` is mounted within
          the container as ``container_path`` with read-write access.
        - ``<host_path>:<container_path>:<selinux_context>`` - ``host_path`` is
          mounted within the container as ``container_path`` with read-write
          access. Additionally, the specified selinux context will be set
          within the container.
        - ``<host_path>:<container_path>:<read_only>`` - ``host_path`` is
          mounted within the container as ``container_path``, with the
          read-only or read-write setting explicitly defined.
        - ``<host_path>:<container_path>:<read_only>,<selinux_context>`` -
          ``host_path`` is mounted within the container as ``container_path``,
          with the read-only or read-write setting explicitly defined.
          Additionally, the specified selinux context will be set within the
          container.

        ``<read_only>`` can be either ``rw`` for read-write access, or ``ro``
        for read-only access. When omitted, it is assumed to be read-write.

        ``<selinux_context>`` can be ``z`` if the volume is shared between
        multiple containers, or ``Z`` if the volume should be private.

        .. note::
            When both ``<read_only>`` and ``<selinux_context>`` are specified,
            there must be a comma before ``<selinux_context>``.

        Binds can be expressed as a comma-separated list or a YAML list. The
        below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - binds: /srv/www:/var/www:ro,/etc/foo.conf:/usr/local/etc/foo.conf:rw

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - binds:
                  - /srv/www:/var/www:ro
                  - /home/myuser/conf/foo.conf:/etc/foo.conf:rw

        However, in cases where both ro/rw and an selinux context are combined,
        the only option is to use a YAML list, like so:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - binds:
                  - /srv/www:/var/www:ro,Z
                  - /home/myuser/conf/foo.conf:/etc/foo.conf:rw,Z

        Since the second bind in the previous example is mounted read-write,
        the ``rw`` and comma can be dropped. For example:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - binds:
                  - /srv/www:/var/www:ro,Z
                  - /home/myuser/conf/foo.conf:/etc/foo.conf:Z

    blkio_weight
        Block IO weight (relative weight), accepts a weight value between 10
        and 1000.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - blkio_weight: 100

    blkio_weight_device
        Block IO weight (relative device weight), specified as a list of
        expressions in the format ``PATH:RATE``

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - blkio_weight_device: /dev/sda:100

    cap_add
        List of capabilities to add within the container. Can be expressed as a
        comma-separated list or a Python list. The below two examples are
        equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cap_add: SYS_ADMIN,MKNOD

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cap_add:
                  - SYS_ADMIN
                  - MKNOD

        .. note::

            This option requires Docker 1.2.0 or newer.

    cap_drop
        List of capabilities to drop within the container. Can be expressed as
        a comma-separated list or a Python list. The below two examples are
        equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cap_drop: SYS_ADMIN,MKNOD

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cap_drop:
                  - SYS_ADMIN
                  - MKNOD

        .. note::
            This option requires Docker 1.2.0 or newer.

    command (or *cmd*)
        Command to run in the container

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - command: bash

    cpuset_cpus (or *cpuset*)
        CPUs on which which to allow execution, specified as a string
        containing a range (e.g. ``0-3``) or a comma-separated list of CPUs
        (e.g. ``0,1``).

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cpuset_cpus: "0,1"

    cpuset_mems
        Memory nodes on which which to allow execution, specified as a string
        containing a range (e.g. ``0-3``) or a comma-separated list of MEMs
        (e.g. ``0,1``). Only effective on NUMA systems.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cpuset_mems: "0,1"

    cpu_group
        The length of a CPU period in microseconds

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cpu_group: 100000

    cpu_period
        Microseconds of CPU time that the container can get in a CPU period

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cpu_period: 50000

    cpu_shares
        CPU shares (relative weight), specified as an integer between 2 and 1024.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - cpu_shares: 512

    detach : False
        If ``True``, run the container's command in the background (daemon
        mode)

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - detach: True

    devices
        List of host devices to expose within the container. Can be expressed
        as a comma-separated list or a YAML list. The below two examples are
        equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices: /dev/net/tun,/dev/xvda1:/dev/xvda1,/dev/xvdb1:/dev/xvdb1:r

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices:
                  - /dev/net/tun
                  - /dev/xvda1:/dev/xvda1
                  - /dev/xvdb1:/dev/xvdb1:r

    device_read_bps
        Limit read rate (bytes per second) from a device, specified as a list
        of expressions in the format ``PATH:RATE``, where ``RATE`` is either an
        integer number of bytes, or a string ending in ``kb``, ``mb``, or
        ``gb``. Can be expressed as a comma-separated list or a YAML list. The
        below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_read_bps: /dev/sda:1mb,/dev/sdb:5mb

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_read_bps:
                  - /dev/sda:1mb
                  - /dev/sdb:5mb

    device_read_iops
        Limit read rate (I/O per second) from a device, specified as a list
        of expressions in the format ``PATH:RATE``, where ``RATE`` is a number
        of I/O operations. Can be expressed as a comma-separated list or a YAML
        list. The below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_read_iops: /dev/sda:1000,/dev/sdb:500

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_read_iops:
                  - /dev/sda:1000
                  - /dev/sdb:500

    device_write_bps
        Limit write rate (bytes per second) from a device, specified as a list
        of expressions in the format ``PATH:RATE``, where ``RATE`` is either an
        integer number of bytes, or a string ending in ``kb``, ``mb``, or
        ``gb``. Can be expressed as a comma-separated list or a YAML list. The
        below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_write_bps: /dev/sda:1mb,/dev/sdb:5mb

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_write_bps:
                  - /dev/sda:1mb
                  - /dev/sdb:5mb


    device_write_iops
        Limit write rate (I/O per second) from a device, specified as a list
        of expressions in the format ``PATH:RATE``, where ``RATE`` is a number
        of I/O operations. Can be expressed as a comma-separated list or a
        YAML list. The below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_write_iops: /dev/sda:1000,/dev/sdb:500

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - devices_write_iops:
                  - /dev/sda:1000
                  - /dev/sdb:500

    dns
        List of DNS nameservers. Can be expressed as a comma-separated list or
        a YAML list. The below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dns: 8.8.8.8,8.8.4.4

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dns:
                  - 8.8.8.8
                  - 8.8.4.4

        .. note::

            To skip IP address validation, use ``validate_ip_addrs=False``

    dns_opt
        Additional options to be added to the container’s ``resolv.conf`` file.
        Can be expressed as a comma-separated list or a YAML list. The below
        two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dns_opt: ndots:9

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dns_opt:
                  - ndots:9

    dns_search
        List of DNS search domains. Can be expressed as a comma-separated list
        or a YAML list. The below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dns_search: foo1.domain.tld,foo2.domain.tld

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dns_search:
                  - foo1.domain.tld
                  - foo2.domain.tld

    domainname
        The domain name to use for the container

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - dommainname: domain.tld

    entrypoint
        Entrypoint for the container

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - entrypoint: "mycmd --arg1 --arg2"

        This argument can also be specified as a list:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - entrypoint:
                  - mycmd
                  - --arg1
                  - --arg2

    environment
        Either a list of variable/value mappings, or a list of strings in the
        format ``VARNAME=value``. The below three examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - environment:
                  - VAR1: value
                  - VAR2: value

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - environment: 'VAR1=value,VAR2=value'

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - environment:
                  - VAR1=value
                  - VAR2=value

    extra_hosts
        Additional hosts to add to the container's /etc/hosts file. Can be
        expressed as a comma-separated list or a Python list. The below two
        examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - extra_hosts: web1:10.9.8.7,web2:10.9.8.8

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - extra_hosts:
                  - web1:10.9.8.7
                  - web2:10.9.8.8

        .. note::

            To skip IP address validation, use ``validate_ip_addrs=False``

        .. note::

            This option requires Docker 1.3.0 or newer.

    group_add
        List of additional group names and/or IDs that the container process
        will run as. Can be expressed as a comma-separated list or a YAML list.
        The below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - group_add: web,network

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - group_add:
                  - web
                  - network

    hostname
        Hostname of the container. If not provided, the value passed as the
        container's``name`` will be used for the hostname.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - hostname: web1

        .. warning::

            ``hostname`` cannot be set if ``network_mode`` is set to ``host``.
            The below example will result in an error:

            .. code-block:: yaml

                foo:
                  docker_container.running:
                    - image: bar/baz:latest
                    - hostname: web1
                    - network_mode: host

    interactive (or *stdin_open*) : False
        Leave stdin open, even if not attached

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - interactive: True

    ipc_mode (or *ipc*)
        Set the IPC mode for the container. The default behavior is to create a
        private IPC namespace for the container, but this option can be
        used to change that behavior:

        - ``container:<container_name_or_id>`` reuses another container shared
          memory, semaphores and message queues
        - ``host``: use the host's shared memory, semaphores and message queues

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ipc_mode: container:foo

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ipc_mode: host

        .. warning::
            Using ``host`` gives the container full access to local shared
            memory and is therefore considered insecure.

    isolation
        Specifies the type of isolation technology used by containers

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - isolation: hyperv

        .. note::
            The default value on Windows server is ``process``, while the
            default value on Windows client is ``hyperv``. On Linux, only
            ``default`` is supported.

    labels
        Add metadata to the container. Labels can be set both with and without
        values, and labels with values can be passed either as ``key=value`` or
        ``key: value`` pairs. For example, while the below would be very
        confusing to read, it is technically valid, and demonstrates the
        different ways in which labels can be passed:

        .. code-block:: yaml

            mynet:
              docker_network.present:
                - labels:
                  - foo
                  - bar=baz
                  - hello: world

        The labels can also simply be passed as a YAML dictionary, though this
        can be error-prone due to some :ref:`idiosyncrasies
        <yaml-idiosyncrasies>` with how PyYAML loads nested data structures:

        .. code-block:: yaml

            foo:
              docker_network.present:
                - labels:
                    foo: ''
                    bar: baz
                    hello: world

        .. versionchanged:: 2018.3.0
            Methods for specifying labels can now be mixed. Earlier releases
            required either labels with or without values.

    links
        Link this container to another. Links can be specified as a list of
        mappings or a comma-separated or Python list of expressions in the
        format ``<container_name_or_id>:<link_alias>``. The below three
        examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - links:
                  - web1: link1
                  - web2: link2

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - links: web1:link1,web2:link2

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - links:
                  - web1:link1
                  - web2:link2

    log_driver and log_opt
        Set container's logging driver and options to configure that driver.
        Requires Docker 1.6 or newer.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - log_driver: syslog
                - log_opt:
                  - syslog-address: tcp://192.168.0.42
                  - syslog-facility: daemon

        The ``log_opt`` can also be expressed as a comma-separated or YAML list
        of ``key=value`` pairs. The below two examples are equivalent to the
        above one:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - log_driver: syslog
                - log_opt: "syslog-address=tcp://192.168.0.42,syslog-facility=daemon"

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - log_driver: syslog
                - log_opt:
                  - syslog-address=tcp://192.168.0.42
                  - syslog-facility=daemon

        .. note::
            The logging driver feature was improved in Docker 1.13 introducing
            option name changes. Please see Docker's
            `Configure logging drivers`_ documentation for more information.

        .. _`Configure logging drivers`: https://docs.docker.com/engine/admin/logging/overview/

    lxc_conf
        Additional LXC configuration parameters to set before starting the
        container. Either a list of variable/value mappings, or a list of
        strings in the format ``VARNAME=value``. The below three examples are
        equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - lxc_conf:
                  - lxc.utsname: docker
                  - lxc.arch: x86_64

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - lxc_conf: lxc.utsname=docker,lxc.arch=x86_64

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - lxc_conf:
                  - lxc.utsname=docker
                  - lxc.arch=x86_64

        .. note::
            These LXC configuration parameters will only have the desired
            effect if the container is using the LXC execution driver, which
            has been deprecated for some time.

    mac_address
        MAC address to use for the container. If not specified, a random MAC
        address will be used.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - mac_address: 01:23:45:67:89:0a

    mem_limit (or *memory*) : 0
        Memory limit. Can be specified in bytes or using single-letter units
        (i.e. ``512M``, ``2G``, etc.). A value of ``0`` (the default) means no
        memory limit.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - mem_limit: 512M

    mem_swappiness
        Tune a container's memory swappiness behavior. Accepts an integer
        between 0 and 100.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - mem_swappiness: 60

    memswap_limit (or *memory_swap*) : -1
        Total memory limit (memory plus swap). Set to ``-1`` to disable swap. A
        value of ``0`` means no swap limit.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - memswap_limit: 1G

    network_disabled : False
        If ``True``, networking will be disabled within the container

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - network_disabled: True

    network_mode : bridge
        One of the following:

        - ``bridge`` - Creates a new network stack for the container on the
          docker bridge
        - ``none`` - No networking (equivalent of the Docker CLI argument
          ``--net=none``). Not to be confused with Python's ``None``.
        - ``container:<name_or_id>`` - Reuses another container's network stack
        - ``host`` - Use the host's network stack inside the container

          .. warning::

                Using ``host`` mode gives the container full access to the
                hosts system's services (such as D-bus), and is therefore
                considered insecure.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - network_mode: "none"

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - network_mode: container:web1

    oom_kill_disable
        Whether to disable OOM killer

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - oom_kill_disable: False

    oom_score_adj
        An integer value containing the score given to the container in order
        to tune OOM killer preferences

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - oom_score_adj: 500

    pid_mode
        Set to ``host`` to use the host container's PID namespace within the
        container. Requires Docker 1.5.0 or newer.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - pid_mode: host

        .. note::
            This option requires Docker 1.5.0 or newer.

    pids_limit
        Set the container's PID limit. Set to ``-1`` for unlimited.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - pids_limit: 2000

    port_bindings (or *publish*)
        Bind exposed ports. Port bindings should be passed in the same way as
        the ``--publish`` argument to the ``docker run`` CLI command:

        - ``ip:hostPort:containerPort`` - Bind a specific IP and port on the
          host to a specific port within the container.
        - ``ip::containerPort`` - Bind a specific IP and an ephemeral port to a
          specific port within the container.
        - ``hostPort:containerPort`` - Bind a specific port on all of the
          host's interfaces to a specific port within the container.
        - ``containerPort`` - Bind an ephemeral port on all of the host's
          interfaces to a specific port within the container.

        Multiple bindings can be separated by commas, or expressed as a YAML
        list, and port ranges can be defined using dashes. The below two
        examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - port_bindings: "4505-4506:14505-14506,2123:2123/udp,8080"

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - port_bindings:
                  - 4505-4506:14505-14506
                  - 2123:2123/udp
                  - 8080

        .. note::
            When specifying a protocol, it must be passed in the
            ``containerPort`` value, as seen in the examples above.

    ports
        A list of ports to expose on the container. Can either be a
        comma-separated list or a YAML list. If the protocol is omitted, the
        port will be assumed to be a TCP port. The below two examples are
        equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ports: 1111,2222/udp

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ports:
                  - 1111
                  - 2222/udp

    privileged : False
        If ``True``, runs the exec process with extended privileges

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - privileged: True

    publish_all_ports (or *publish_all*) : False
        Publish all ports to the host

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ports: 8080
                - publish_all_ports: True

    read_only : False
        If ``True``, mount the container’s root filesystem as read only

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - read_only: True

    restart_policy (or *restart*)
        Set a restart policy for the container. Must be passed as a string in
        the format ``policy[:retry_count]`` where ``policy`` is one of
        ``always``, ``unless-stopped``, or ``on-failure``, and ``retry_count``
        is an optional limit to the number of retries. The retry count is ignored
        when using the ``always`` or ``unless-stopped`` restart policy.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - restart_policy: on-failure:5

            bar:
              docker_container.running:
                - image: bar/baz:latest
                - restart_policy: always

    security_opt (or *security_opts*):
        Security configuration for MLS systems such as SELinux and AppArmor.
        Can be expressed as a comma-separated list or a YAML list. The below
        two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - security_opt: apparmor:unconfined

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - security_opt:
                  - apparmor:unconfined

        .. important::
            Some security options can contain commas. In these cases, this
            argument *must* be passed as a Python list, as splitting by comma
            will result in an invalid configuration.

        .. note::
            See the documentation for security_opt at
            https://docs.docker.com/engine/reference/run/#security-configuration

    shm_size
        Size of /dev/shm

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - shm_size: 128M

    stop_signal
        Specify the signal docker will send to the container when stopping.
        Useful when running systemd as PID 1 inside the container.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - stop_signal: SIGRTMIN+3

        .. note::

            This option requires Docker 1.9.0 or newer and docker-py 1.7.0 or
            newer.

        .. versionadded:: 2016.11.0

    stop_timeout
        Timeout to stop the container, in seconds

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - stop_timeout: 5

        .. note::
            In releases prior to 2017.7.0, this option was not set in the
            container configuration, but rather this timeout was enforced only
            when shutting down an existing container to replace it. To remove
            the ambiguity, and to allow for the container to have a stop
            timeout set for it, the old ``stop_timeout`` argument has been
            renamed to ``shutdown_timeout``, while ``stop_timeout`` now refer's
            to the container's configured stop timeout.

    storage_opt
        Storage driver options for the container. Can be either a list of
        strings in the format ``option=value``, or a list of mappings between
        option and value. The below three examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - storage_opt:
                  - dm.basesize: 40G

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - storage_opt: dm.basesize=40G

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - storage_opt:
                  - dm.basesize=40G

    sysctls (or *sysctl*)
        Set sysctl options for the container. Can be either a list of strings
        in the format ``option=value``, or a list of mappings between option
        and value. The below three examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - sysctls:
                  - fs.nr_open: 1048576
                  - kernel.pid_max: 32768

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - sysctls: fs.nr_open=1048576,kernel.pid_max=32768

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - sysctls:
                  - fs.nr_open=1048576
                  - kernel.pid_max=32768

    tmpfs
        A map of container directories which should be replaced by tmpfs mounts
        and their corresponding mount options.

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - tmpfs:
                  - /run: rw,noexec,nosuid,size=65536k

    tty : False
        Attach TTYs

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - tty: True

    ulimits
        List of ulimits. These limits should be passed in the format
        ``<ulimit_name>:<soft_limit>:<hard_limit>``, with the hard limit being
        optional. Can be expressed as a comma-separated list or a YAML list.
        The below two examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ulimits: nofile=1024:1024,nproc=60

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - ulimits:
                  - nofile=1024:1024
                  - nproc=60

    user
        User under which to run exec process

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - user: foo

    userns_mode (or *user_ns_mode*)
        Sets the user namsepace mode, when the user namespace remapping option
        is enabled

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - userns_mode: host

    volumes (or *volume*)
        List of directories to expose as volumes. Can be expressed as a
        comma-separated list or a YAML list. The below two examples are
        equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - volumes: /mnt/vol1,/mnt/vol2

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - volumes:
                  - /mnt/vol1
                  - /mnt/vol2

    volumes_from
        Container names or IDs from which the container will get volumes. Can
        be expressed as a comma-separated list or a YAML list. The below two
        examples are equivalent:

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - volumes_from: foo

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - volumes_from:
                  - foo

    volume_driver
        sets the container's volume driver

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - volume_driver: foobar

    working_dir (or *workdir*)
        Working directory inside the container

        .. code-block:: yaml

            foo:
              docker_container.running:
                - image: bar/baz:latest
                - working_dir: /var/log/nginx
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if image is None:
        ret["result"] = False
        ret["comment"] = "The 'image' argument is required"
        return ret
    elif not isinstance(image, str):
        image = str(image)

    try:
        # Since we're rewriting the "networks" value below, save the original
        # value here.
        configured_networks = networks
        networks = _parse_networks(networks)
        if networks:
            kwargs["networks"] = networks
        image_id = _resolve_image(ret, image, client_timeout)
    except CommandExecutionError as exc:
        ret["result"] = False
        if exc.info is not None:
            return _format_comments(ret, exc.info)
        else:
            ret["comment"] = str(exc)
            return ret

    comments = []

    # Pop off the send_signal argument passed by the watch requisite
    send_signal = kwargs.pop("send_signal", False)

    try:
        current_image_id = __salt__["docker.inspect_container"](name)["Image"]
    except CommandExecutionError:
        current_image_id = None
    except KeyError:
        ret["result"] = False
        comments.append(
            "Unable to detect current image for container '{}'. "
            "This might be due to a change in the Docker API.".format(name)
        )
        return _format_comments(ret, comments)

    # Shorthand to make the below code more understandable
    exists = current_image_id is not None

    pre_state = __salt__["docker.state"](name) if exists else None

    # If skip_comparison is True, we're definitely going to be using the temp
    # container as the new container (because we're forcing the change, or
    # because the image IDs differ). If False, we'll need to perform a
    # comparison between it and the new container.
    skip_comparison = force or not exists or current_image_id != image_id

    if skip_comparison and __opts__["test"]:
        ret["result"] = None
        if force:
            ret["changes"]["forced_update"] = True
        elif current_image_id != image_id:
            ret["changes"]["image"] = {"old": current_image_id, "new": image_id}
        comments.append(
            "Container '{}' would be {}".format(
                name, "created" if not exists else "replaced"
            )
        )
        return _format_comments(ret, comments)

    # Create temp container (or just create the named container if the
    # container does not already exist)
    try:
        temp_container = __salt__["docker.create"](
            image,
            name=name if not exists else None,
            skip_translate=skip_translate,
            ignore_collisions=ignore_collisions,
            validate_ip_addrs=validate_ip_addrs,
            client_timeout=client_timeout,
            **kwargs,
        )
        temp_container_name = temp_container["Name"]
    except KeyError as exc:
        ret["result"] = False
        comments.append(
            "Key '{}' missing from API response, this may be due to a "
            "change in the Docker Remote API. Please report this on the "
            "SaltStack issue tracker if it has not already been reported.".format(exc)
        )
        return _format_comments(ret, comments)
    except Exception as exc:  # pylint: disable=broad-except
        ret["result"] = False
        msg = str(exc)
        if (
            isinstance(exc, CommandExecutionError)
            and isinstance(exc.info, dict)
            and "invalid" in exc.info
        ):
            msg += (
                "\n\nIf you feel this information is incorrect, the "
                "skip_translate argument can be used to skip input "
                "translation for the argument(s) identified as invalid. See "
                "the documentation for details."
            )
        comments.append(msg)
        return _format_comments(ret, comments)

    def _replace(orig, new):
        rm_kwargs = {"stop": True}
        if shutdown_timeout is not None:
            rm_kwargs["timeout"] = shutdown_timeout
        ret["changes"].setdefault("container_id", {})["removed"] = __salt__[
            "docker.rm"
        ](name, **rm_kwargs)
        try:
            result = __salt__["docker.rename"](new, orig)
        except CommandExecutionError as exc:
            result = False
            comments.append(f"Failed to rename temp container: {exc}")
        if result:
            comments.append(f"Replaced container '{orig}'")
        else:
            comments.append("Failed to replace container '{0}'")
        return result

    def _delete_temp_container():
        log.debug("Removing temp container '%s'", temp_container_name)
        __salt__["docker.rm"](temp_container_name)

    # If we're not skipping the comparison, then the assumption is that
    # temp_container will be discarded, unless the comparison reveals
    # differences, in which case we'll set cleanup_temp = False to prevent it
    # from being cleaned.
    cleanup_temp = not skip_comparison
    try:
        pre_net_connect = __salt__["docker.inspect_container"](
            name if exists else temp_container_name
        )
        for net_name, net_conf in networks.items():
            try:
                __salt__["docker.connect_container_to_network"](
                    temp_container_name, net_name, **net_conf
                )
            except CommandExecutionError as exc:
                # Shouldn't happen, stopped docker containers can be
                # attached to networks even if the static IP lies outside
                # of the network's subnet. An exception will be raised once
                # you try to start the container, however.
                ret["result"] = False
                comments.append(str(exc))
                return _format_comments(ret, comments)

        post_net_connect = __salt__["docker.inspect_container"](temp_container_name)

        if configured_networks is not None:
            # Use set arithmetic to determine the networks which are connected
            # but not explicitly defined. They will be disconnected below. Note
            # that we check configured_networks because it represents the
            # original (unparsed) network configuration. When no networks
            # argument is used, the parsed networks will be an empty list, so
            # it's not sufficient to do a boolean check on the "networks"
            # variable.
            extra_nets = set(
                post_net_connect.get("NetworkSettings", {}).get("Networks", {})
            ) - set(networks)

            if extra_nets:
                for extra_net in extra_nets:
                    __salt__["docker.disconnect_container_from_network"](
                        temp_container_name, extra_net
                    )

                # We've made changes, so we need to inspect the container again
                post_net_connect = __salt__["docker.inspect_container"](
                    temp_container_name
                )

        net_changes = __salt__["docker.compare_container_networks"](
            pre_net_connect, post_net_connect
        )

        if not skip_comparison:
            docker_version_info = __salt__["docker.version"]()["VersionInfo"]
            if docker_version_info < (25, 0):
                compare_containers_ignore = "Hostname"
            else:
                # With docker >= 25.0 we get a new value to compare,
                # MacAddress, which we'll ignore for now.
                compare_containers_ignore = "Hostname,MacAddress"
            container_changes = __salt__["docker.compare_containers"](
                name,
                temp_container_name,
                ignore=compare_containers_ignore,
            )
            if container_changes:
                if _check_diff(container_changes):
                    ret.setdefault("warnings", []).append(
                        "The detected changes may be due to incorrect "
                        "handling of arguments in earlier Salt releases. If "
                        "this warning persists after running the state "
                        "again{}, and no changes were made to the SLS file, "
                        "then please report this.".format(
                            " without test=True" if __opts__["test"] else ""
                        )
                    )

                changes_ptr = ret["changes"].setdefault("container", {})
                changes_ptr.update(container_changes)
                if __opts__["test"]:
                    ret["result"] = None
                    comments.append(
                        "Container '{}' would be {}".format(
                            name, "created" if not exists else "replaced"
                        )
                    )
                else:
                    # We don't want to clean the temp container, we'll be
                    # replacing the existing one with it.
                    cleanup_temp = False
                    # Replace the container
                    if not _replace(name, temp_container_name):
                        ret["result"] = False
                        return _format_comments(ret, comments)
                    ret["changes"].setdefault("container_id", {})["added"] = (
                        temp_container["Id"]
                    )
            else:
                # No changes between existing container and temp container.
                # First check if a requisite is asking to send a signal to the
                # existing container.
                if send_signal:
                    if __opts__["test"]:
                        comments.append(
                            f"Signal {watch_action} would be sent to container"
                        )
                    else:
                        try:
                            __salt__["docker.signal"](name, signal=watch_action)
                        except CommandExecutionError as exc:
                            ret["result"] = False
                            comments.append(f"Failed to signal container: {exc}")
                            return _format_comments(ret, comments)
                        else:
                            ret["changes"]["signal"] = watch_action
                            comments.append(f"Sent signal {watch_action} to container")
                elif container_changes:
                    if not comments:
                        log.warning(
                            "docker_container.running: detected changes without "
                            "a specific comment for container '%s'",
                            name,
                        )
                        comments.append(
                            "Container '{}'{} updated.".format(
                                name, " would be" if __opts__["test"] else ""
                            )
                        )
                else:
                    # Container was not replaced, no differences between the
                    # existing container and the temp container were detected,
                    # and no signal was sent to the container.
                    comments.append(
                        f"Container '{name}' is already configured as specified"
                    )

        if net_changes:
            ret["changes"].setdefault("container", {})["Networks"] = net_changes
            if __opts__["test"]:
                ret["result"] = None
                comments.append("Network configuration would be updated")
            elif cleanup_temp:
                # We only need to make network changes if the container
                # isn't being replaced, since we would already have
                # attached all the networks for purposes of comparison.
                network_failure = False
                for net_name in sorted(net_changes):
                    errors = []
                    disconnected = connected = False
                    try:
                        if name in __salt__["docker.connected"](net_name):
                            __salt__["docker.disconnect_container_from_network"](
                                name, net_name
                            )
                            disconnected = True
                    except CommandExecutionError as exc:
                        errors.append(str(exc))

                    if net_name in networks:
                        try:
                            __salt__["docker.connect_container_to_network"](
                                name, net_name, **networks[net_name]
                            )
                            connected = True
                        except CommandExecutionError as exc:
                            errors.append(str(exc))
                            if disconnected:
                                # We succeeded in disconnecting but failed
                                # to reconnect. This can happen if the
                                # network's subnet has changed and we try
                                # to reconnect with the same IP address
                                # from the old subnet.
                                for item in list(net_changes[net_name]):
                                    if net_changes[net_name][item]["old"] is None:
                                        # Since they'd both be None, just
                                        # delete this key from the changes
                                        del net_changes[net_name][item]
                                    else:
                                        net_changes[net_name][item]["new"] = None

                    if errors:
                        comments.extend(errors)
                        network_failure = True

                    ret["changes"].setdefault("container", {}).setdefault(
                        "Networks", {}
                    )[net_name] = net_changes[net_name]

                    if disconnected and connected:
                        comments.append(
                            "Reconnected to network '{}' with updated "
                            "configuration".format(net_name)
                        )
                    elif disconnected:
                        comments.append(f"Disconnected from network '{net_name}'")
                    elif connected:
                        comments.append(f"Connected to network '{net_name}'")

                if network_failure:
                    ret["result"] = False
                    return _format_comments(ret, comments)
    finally:
        if cleanup_temp:
            _delete_temp_container()

    if skip_comparison:
        if not exists:
            comments.append(f"Created container '{name}'")
        else:
            if not _replace(name, temp_container):
                ret["result"] = False
                return _format_comments(ret, comments)
        ret["changes"].setdefault("container_id", {})["added"] = temp_container["Id"]

    # "exists" means that a container by the specified name existed prior to
    #     this state being run
    # "not cleanup_temp" means that the temp container became permanent, either
    #     because the named container did not exist or changes were detected
    # "cleanup_temp" means that the container already existed and no changes
    #     were detected, so the temp container was discarded
    if (
        not cleanup_temp
        and (not exists or (exists and start))
        or (start and cleanup_temp and pre_state != "running")
    ):
        if __opts__["test"]:
            ret["result"] = None
            comments.append("Container would be started")
            return _format_comments(ret, comments)
        else:
            try:
                post_state = __salt__["docker.start"](name)["state"]["new"]
            except Exception as exc:  # pylint: disable=broad-except
                ret["result"] = False
                comments.append(f"Failed to start container '{name}': '{exc}'")
                return _format_comments(ret, comments)
    else:
        post_state = __salt__["docker.state"](name)

    if not __opts__["test"] and post_state == "running":
        # Now that we're certain the container is running, check each modified
        # network to see if the network went from static (or disconnected) to
        # automatic IP configuration. If so, grab the automatically-assigned
        # IPs and munge the changes dict to include them. Note that this can
        # only be done after the container is started bceause automatic IPs are
        # assigned at runtime.
        contextkey = ".".join((name, "docker_container.running"))

        def _get_nets():
            if contextkey not in __context__:
                new_container_info = __salt__["docker.inspect_container"](name)
                __context__[contextkey] = new_container_info.get(
                    "NetworkSettings", {}
                ).get("Networks", {})
            return __context__[contextkey]

        autoip_keys = __salt__["config.option"](
            "docker.compare_container_networks"
        ).get("automatic", [])
        for net_name, net_changes in (
            ret["changes"].get("container", {}).get("Networks", {}).items()
        ):
            if (
                "IPConfiguration" in net_changes
                and net_changes["IPConfiguration"]["new"] == "automatic"
            ):
                for key in autoip_keys:
                    val = _get_nets().get(net_name, {}).get(key)
                    if val:
                        net_changes[key] = {"old": None, "new": val}
                        try:
                            net_changes.pop("IPConfiguration")
                        except KeyError:
                            pass
        __context__.pop(contextkey, None)

    if pre_state != post_state:
        ret["changes"]["state"] = {"old": pre_state, "new": post_state}
        if pre_state is not None:
            comments.append(f"State changed from '{pre_state}' to '{post_state}'")

    if exists and current_image_id != image_id:
        comments.append("Container has a new image")
        ret["changes"]["image"] = {"old": current_image_id, "new": image_id}

    if post_state != "running" and start:
        ret["result"] = False
        comments.append("Container is not running")

    return _format_comments(ret, comments)


def run(
    name,
    image=None,
    bg=False,
    failhard=True,
    replace=False,
    force=False,
    skip_translate=None,
    ignore_collisions=False,
    validate_ip_addrs=True,
    client_timeout=salt.utils.dockermod.CLIENT_TIMEOUT,
    **kwargs,
):
    """
    .. versionadded:: 2018.3.0

    .. note::
        If no tag is specified in the image name, and nothing matching the
        specified image is pulled on the minion, the ``docker pull`` that
        retrieves the image will pull *all tags* for the image. A tag of
        ``latest`` is not implicit for the pull. For this reason, it is
        recommended to specify the image in ``repo:tag`` notation.

    Like the :py:func:`cmd.run <salt.states.cmd.run>` state, only for Docker.
    Does the equivalent of a ``docker run`` and returns information about the
    container that was created, as well as its output.

    This state accepts the same arguments as :py:func:`docker_container.running
    <salt.states.docker_container.running>`, with the exception of
    ``watch_action``, ``start``, and ``shutdown_timeout`` (though the ``force``
    argument has a different meaning in this state).

    In addition, this state accepts the arguments from :py:func:`docker.logs
    <salt.modules.dockermod.logs>`, with the exception of ``follow``, to
    control how logs are returned.

    Additionally, the following arguments are supported:

    creates
        A path or list of paths. Only run if one or more of the specified paths
        do not exist on the minion.

    bg : False
        If ``True``, run container in background and do not await or deliver
        its results.

        .. note::
            This may not be useful in cases where other states depend on the
            results of this state. Also, the logs will be inaccessible once the
            container exits if ``auto_remove`` is set to ``True``, so keep this
            in mind.

    failhard : True
        If ``True``, the state will return a ``False`` result if the exit code
        of the container is non-zero. When this argument is set to ``False``,
        the state will return a ``True`` result regardless of the container's
        exit code.

        .. note::
            This has no effect if ``bg`` is set to ``True``.

    replace : False
        If ``True``, and if the named container already exists, this will
        remove the existing container. The default behavior is to return a
        ``False`` result when the container already exists.

    force : False
        If ``True``, and the named container already exists, *and* ``replace``
        is also set to ``True``, then the container will be forcibly removed.
        Otherwise, the state will not proceed and will return a ``False``
        result.

    CLI Examples:

    .. code-block:: bash

        salt myminion docker.run_container myuser/myimage command=/usr/local/bin/myscript.sh

    **USAGE EXAMPLE**

    .. code-block:: jinja

        {% set pkg_version = salt.pillar.get('pkg_version', '1.0-1') %}
        build_package:
          docker_container.run:
            - image: myuser/builder:latest
            - binds: /home/myuser/builds:/build_dir
            - command: /scripts/build.sh {{ pkg_version }}
            - creates: /home/myuser/builds/myapp-{{ pkg_version }}.noarch.rpm
            - replace: True
            - networks:
              - mynet
            - require:
              - docker_network: mynet
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    for unsupported in ("watch_action", "start", "shutdown_timeout", "follow"):
        if unsupported in kwargs:
            ret["result"] = False
            ret["comment"] = f"The '{unsupported}' argument is not supported"
            return ret

    if image is None:
        ret["result"] = False
        ret["comment"] = "The 'image' argument is required"
        return ret
    elif not isinstance(image, str):
        image = str(image)

    try:
        if "networks" in kwargs and kwargs["networks"] is not None:
            kwargs["networks"] = _parse_networks(kwargs["networks"])
        _resolve_image(ret, image, client_timeout)
    except CommandExecutionError as exc:
        ret["result"] = False
        if exc.info is not None:
            return _format_comments(ret, exc.info)
        else:
            ret["comment"] = str(exc)
            return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Container would be run{}".format(
            " in the background" if bg else ""
        )
        return ret

    if bg:
        remove = False
    else:
        # We're doing a bit of a hack here, so that we can get the exit code after
        # the container exits. Since the input translation and compilation of the
        # host_config take place within a private function of the execution module,
        # we manually do the handling for auto_remove here and extract if (if
        # present) from the kwargs. This allows us to explicitly pass auto_remove
        # as False when we run the container, so it is still present upon exit (and
        # the exit code can be retrieved). We can then remove the container
        # manually if auto_remove is True.
        remove = None
        for item in ("auto_remove", "rm"):
            try:
                val = kwargs.pop(item)
            except KeyError:
                continue
            if remove is not None:
                if not ignore_collisions:
                    ret["result"] = False
                    ret["comment"] = (
                        "'rm' is an alias for 'auto_remove', they cannot both be used"
                    )
                    return ret
            else:
                remove = bool(val)

        if remove is not None:
            # We popped off the value, so replace it with False
            kwargs["auto_remove"] = False
        else:
            remove = False

    try:
        ret["changes"] = __salt__["docker.run_container"](
            image,
            name=name,
            skip_translate=skip_translate,
            ignore_collisions=ignore_collisions,
            validate_ip_addrs=validate_ip_addrs,
            client_timeout=client_timeout,
            bg=bg,
            replace=replace,
            force=force,
            **kwargs,
        )
    except Exception as exc:  # pylint: disable=broad-except
        log.exception("Encountered error running container")
        ret["result"] = False
        ret["comment"] = f"Encountered error running container: {exc}"
    else:
        if bg:
            ret["comment"] = "Container was run in the background"
        else:
            try:
                retcode = ret["changes"]["ExitCode"]
            except KeyError:
                pass
            else:
                ret["result"] = False if failhard and retcode != 0 else True
                ret["comment"] = (
                    f"Container ran and exited with a return code of {retcode}"
                )

    if remove:
        id_ = ret.get("changes", {}).get("Id")
        if id_:
            try:
                __salt__["docker.rm"](ret["changes"]["Id"])
            except CommandExecutionError as exc:
                ret.setdefault("warnings", []).append(
                    f"Failed to auto_remove container: {exc}"
                )

    return ret


def stopped(
    name=None,
    containers=None,
    shutdown_timeout=None,
    unpause=False,
    error_on_absent=True,
    **kwargs,
):
    """
    Ensure that a container (or containers) is stopped

    name
        Name or ID of the container

    containers
        Run this state on more than one container at a time. The following two
        examples accomplish the same thing:

        .. code-block:: yaml

            stopped_containers:
              docker_container.stopped:
                - names:
                  - foo
                  - bar
                  - baz

        .. code-block:: yaml

            stopped_containers:
              docker_container.stopped:
                - containers:
                  - foo
                  - bar
                  - baz

        However, the second example will be a bit quicker since Salt will stop
        all specified containers in a single run, rather than executing the
        state separately on each image (as it would in the first example).

    shutdown_timeout
        Timeout for graceful shutdown of the container. If this timeout is
        exceeded, the container will be killed. If this value is not passed,
        then the container's configured ``stop_timeout`` will be observed. If
        ``stop_timeout`` was also unset on the container, then a timeout of 10
        seconds will be used.

    unpause : False
        Set to ``True`` to unpause any paused containers before stopping. If
        unset, then an error will be raised for any container that was paused.

    error_on_absent : True
        By default, this state will return an error if any of the specified
        containers are absent. Set this to ``False`` to suppress that error.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if not name and not containers:
        ret["comment"] = "One of 'name' and 'containers' must be provided"
        return ret
    if containers is not None:
        if not isinstance(containers, list):
            ret["comment"] = "containers must be a list"
            return ret
        targets = []
        for target in containers:
            if not isinstance(target, str):
                target = str(target)
            targets.append(target)
    elif name:
        if not isinstance(name, str):
            targets = [str(name)]
        else:
            targets = [name]

    containers = {}
    for target in targets:
        try:
            c_state = __salt__["docker.state"](target)
        except CommandExecutionError:
            containers.setdefault("absent", []).append(target)
        else:
            containers.setdefault(c_state, []).append(target)

    errors = []
    if error_on_absent and "absent" in containers:
        errors.append(
            "The following container(s) are absent: {}".format(
                ", ".join(containers["absent"])
            )
        )

    if not unpause and "paused" in containers:
        ret["result"] = False
        errors.append(
            "The following container(s) are paused: {}".format(
                ", ".join(containers["paused"])
            )
        )

    if errors:
        ret["result"] = False
        ret["comment"] = ". ".join(errors)
        return ret

    to_stop = containers.get("running", []) + containers.get("paused", [])

    if not to_stop:
        ret["result"] = True
        if len(targets) == 1:
            ret["comment"] = f"Container '{targets[0]}' is "
        else:
            ret["comment"] = "All specified containers are "
        if "absent" in containers:
            ret["comment"] += "absent or "
        ret["comment"] += "not running"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "The following container(s) will be stopped: {}".format(
            ", ".join(to_stop)
        )
        return ret

    stop_errors = []
    for target in to_stop:
        stop_kwargs = {"unpause": unpause}
        if shutdown_timeout:
            stop_kwargs["timeout"] = shutdown_timeout
        changes = __salt__["docker.stop"](target, **stop_kwargs)
        if changes["result"] is True:
            ret["changes"][target] = changes
        else:
            if "comment" in changes:
                stop_errors.append(changes["comment"])
            else:
                stop_errors.append(f"Failed to stop container '{target}'")

    if stop_errors:
        ret["comment"] = "; ".join(stop_errors)
        return ret

    ret["result"] = True
    ret["comment"] = "The following container(s) were stopped: {}".format(
        ", ".join(to_stop)
    )
    return ret


def absent(name, force=False):
    """
    Ensure that a container is absent

    name
        Name of the container

    force : False
        Set to ``True`` to remove the container even if it is running

    Usage Examples:

    .. code-block:: yaml

        mycontainer:
          docker_container.absent

        multiple_containers:
          docker_container.absent:
            - names:
              - foo
              - bar
              - baz
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if name not in __salt__["docker.list_containers"](all=True):
        ret["result"] = True
        ret["comment"] = f"Container '{name}' does not exist"
        return ret

    pre_state = __salt__["docker.state"](name)
    if pre_state != "stopped" and not force:
        ret["comment"] = "Container is running, set force to True to forcibly remove it"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Container '{name}' will be removed"
        return ret

    try:
        ret["changes"]["removed"] = __salt__["docker.rm"](name, force=force)
    except Exception as exc:  # pylint: disable=broad-except
        ret["comment"] = f"Failed to remove container '{name}': {exc}"
        return ret

    if name in __salt__["docker.list_containers"](all=True):
        ret["comment"] = f"Failed to remove container '{name}'"
    else:
        if force and pre_state != "stopped":
            method = "Forcibly"
        else:
            method = "Successfully"
        ret["comment"] = f"{method} removed container '{name}'"
        ret["result"] = True
    return ret


def mod_watch(name, sfun=None, **kwargs):
    """
    The docker_container watcher, called to invoke the watch command.

    .. note::
        This state exists to support special handling of the ``watch``
        :ref:`requisite <requisites>`. It should not be called directly.

        Parameters for this function should be set by the state being triggered.
    """
    if sfun == "running":
        watch_kwargs = copy.deepcopy(kwargs)
        if watch_kwargs.get("watch_action", "force") == "force":
            watch_kwargs["force"] = True
        else:
            watch_kwargs["send_signal"] = True
            watch_kwargs["force"] = False
        return running(name, **watch_kwargs)

    if sfun == "stopped":
        return stopped(name, **salt.utils.args.clean_kwargs(**kwargs))

    if sfun == "run":
        return run(name, **salt.utils.args.clean_kwargs(**kwargs))

    return {
        "name": name,
        "changes": {},
        "result": False,
        "comment": f"watch requisite is not implemented for {sfun}",
    }



from pyroute2.bsd.pf_route import (
    if_announcemsg,
    if_msg,
    ifa_msg,
    ifma_msg,
    rt_msg,
)

RTM_ADD = 0x1  # Add Route
RTM_DELETE = 0x2  # Delete Route
RTM_CHANGE = 0x3  # Change Metrics or flags
RTM_GET = 0x4  # Report Metrics
RTM_LOSING = 0x5  # Kernel Suspects Partitioning
RTM_REDIRECT = 0x6  # Told to use different route
RTM_MISS = 0x7  # Lookup failed on this address
RTM_LOCK = 0x8  # Fix specified metrics
RTM_RESOLVE = 0xB  # Req to resolve dst to LL addr
RTM_NEWADDR = 0xC  # Address being added to iface
RTM_DELADDR = 0xD  # Address being removed from iface
RTM_IFINFO = 0xE  # Iface going up/down etc
RTM_NEWMADDR = 0xF  # Mcast group membership being added to if
RTM_DELMADDR = 0x10  # Mcast group membership being deleted
RTM_IFANNOUNCE = 0x11  # Iface arrival/departure
RTM_IEEE80211 = 0x12  # IEEE80211 wireless event


class RTMSocketBase(object):
    msg_map = {
        RTM_ADD: rt_msg,
        RTM_DELETE: rt_msg,
        RTM_CHANGE: rt_msg,
        RTM_GET: rt_msg,
        RTM_LOSING: rt_msg,
        RTM_REDIRECT: rt_msg,
        RTM_MISS: rt_msg,
        RTM_LOCK: rt_msg,
        RTM_RESOLVE: rt_msg,
        RTM_NEWADDR: ifa_msg,
        RTM_DELADDR: ifa_msg,
        RTM_IFINFO: if_msg,
        RTM_NEWMADDR: ifma_msg,
        RTM_DELMADDR: ifma_msg,
        RTM_IFANNOUNCE: if_announcemsg,
        RTM_IEEE80211: if_announcemsg,
    }



"""
Vault SDB Module

:maintainer:    SaltStack
:maturity:      New
:platform:      all

.. versionadded:: 2016.11.0

This module allows access to Hashicorp Vault using an ``sdb://`` URI.

Base configuration instructions are documented in the :ref:`execution module docs <vault-setup>`.
Below are noted extra configuration required for the sdb module, but the base
configuration must also be completed.

Like all sdb modules, the vault module requires a configuration profile to
be configured in either the minion configuration file or a pillar. This profile
requires only setting the ``driver`` parameter to ``vault``:

.. code-block:: yaml

    myvault:
      driver: vault

Once configured you can access data using a URL such as:

.. code-block:: yaml

    password: sdb://myvault/secret/passwords/mypassword

In this URL, ``myvault`` refers to the configuration profile,
``secret/passwords`` is the path where the data resides, and ``mypassword`` is
the key of the data to return.

The above URI is analogous to running the following vault command:

.. code-block:: bash

    $ vault read -field=mypassword secret/passwords


Further configuration
---------------------
The following options can be set in the profile:

patch
    When writing data, partially update the secret instead of overwriting it completely.
    This is usually the expected behavior, since without this option,
    each secret path can only contain a single mapping key safely.
    Defaults to ``False`` for backwards-compatibility reasons.

    .. versionadded:: 3007.0
"""

import logging

import salt.exceptions
import salt.utils.vault as vault

log = logging.getLogger(__name__)

__func_alias__ = {"set_": "set"}


def set_(key, value, profile=None):
    """
    Set a key/value pair in the vault service
    """
    if "?" in key:
        path, key = key.split("?")
    else:
        path, key = key.rsplit("/", 1)
    data = {key: value}
    curr_data = {}
    profile = profile or {}

    if profile.get("patch"):
        try:
            # Patching only works on existing secrets.
            # Save the current data if patching is enabled
            # to write it back later, if any errors happen in patch_kv.
            # This also checks that the path exists, otherwise patching fails as well.
            curr_data = vault.read_kv(path, __opts__, __context__)
            vault.patch_kv(path, data, __opts__, __context__)
            return True
        except (vault.VaultNotFoundError, vault.VaultPermissionDeniedError):
            pass

    curr_data.update(data)
    try:
        vault.write_kv(path, data, __opts__, __context__)
        return True
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to write secret! %s: %s", type(err).__name__, err)
        raise salt.exceptions.CommandExecutionError(err) from err


def get(key, profile=None):
    """
    Get a value from the vault service
    """
    full_path = key
    if "?" in key:
        path, key = key.split("?")
    else:
        path, key = key.rsplit("/", 1)

    try:
        try:
            res = vault.read_kv(path, __opts__, __context__)
            if key in res:
                return res[key]
            return None
        except vault.VaultNotFoundError:
            return vault.read_kv(full_path, __opts__, __context__)
    except vault.VaultNotFoundError:
        return None
    except Exception as err:  # pylint: disable=broad-except
        log.error("Failed to read secret! %s: %s", type(err).__name__, err)
        raise salt.exceptions.CommandExecutionError(err) from err



"""
Manage Apache Modules

.. versionadded:: 2014.7.0

Enable and disable apache modules.

.. code-block:: yaml

    Enable cgi module:
      apache_module.enabled:
        - name: cgi

    Disable cgi module:
      apache_module.disabled:
        - name: cgi
"""

__deprecated__ = (
    3009,
    "apache",
    "https://github.com/salt-extensions/saltext-apache",
)


def __virtual__():
    """
    Only load if a2enmod is available.
    """
    if "apache.a2enmod" in __salt__:
        return "apache_module"
    return (False, "apache module could not be loaded")


def enabled(name):
    """
    Ensure an Apache module is enabled.

    .. versionadded:: 2016.3.0

    name
        Name of the Apache module
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_enabled = __salt__["apache.check_mod_enabled"](name)
    if not is_enabled:
        if __opts__["test"]:
            msg = f"Apache module {name} is set to be enabled."
            ret["comment"] = msg
            ret["changes"]["old"] = None
            ret["changes"]["new"] = name
            ret["result"] = None
            return ret
        status = __salt__["apache.a2enmod"](name)["Status"]
        if isinstance(status, str) and "enabled" in status:
            ret["result"] = True
            ret["changes"]["old"] = None
            ret["changes"]["new"] = name
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to enable {name} Apache module"
            if isinstance(status, str):
                ret["comment"] = ret["comment"] + f" ({status})"
            return ret
    else:
        ret["comment"] = f"{name} already enabled."
    return ret


def disabled(name):
    """
    Ensure an Apache module is disabled.

    .. versionadded:: 2016.3.0

    name
        Name of the Apache module
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_enabled = __salt__["apache.check_mod_enabled"](name)
    if is_enabled:
        if __opts__["test"]:
            msg = f"Apache module {name} is set to be disabled."
            ret["comment"] = msg
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
            ret["result"] = None
            return ret
        status = __salt__["apache.a2dismod"](name)["Status"]
        if isinstance(status, str) and "disabled" in status:
            ret["result"] = True
            ret["changes"]["old"] = name
            ret["changes"]["new"] = None
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to disable {name} Apache module"
            if isinstance(status, str):
                ret["comment"] = ret["comment"] + f" ({status})"
            return ret
    else:
        ret["comment"] = f"{name} already disabled."
    return ret



"""Functions for obtaining CLI input"""

import termios
import signal
import sys
import os
from typing import Union
from . import color


def get_keypress() -> str:
    """Get a single keypress, the ugly POSIX way"""
    fileno = sys.stdin.fileno()
    old = termios.tcgetattr(fileno)
    new = termios.tcgetattr(fileno)
    new[3] = new[3] & ~termios.ICANON & ~termios.ECHO
    new[6][termios.VMIN] = 1
    new[6][termios.VTIME] = 0
    termios.tcsetattr(fileno, termios.TCSANOW, new)
    key = None
    try:
        key = str(os.read(fileno, 4), 'ascii')
    finally:
        termios.tcsetattr(fileno, termios.TCSAFLUSH, old)
    keylist = {
        '\t': 'TAB',
        '\n': 'ENTER',
        '\x7f': 'BACKSPACE',
        '\x1b': 'ESC',
        '\x1b[Z': 'BACKTAB',
        '\x1bOH': 'HOME',
        '\x1bOP': 'F1',
        '\x1bOQ': 'F2',
        '\x1bOR': 'F3',
        '\x1bOS': 'F4',
        '\x1b[15': 'F5',
        '\x1b[17': 'F6',
        '\x1b[18': 'F7',
        '\x1b[19': 'F8',
        '\x1b[20': 'F9',
        '\x1b[21': 'F10',
        '\x1b[23': 'F11',
        '\x1b[24': 'F12',
        '\x1b[A': 'ARROW_UP',
        '\x1b[B': 'ARROW_DN',
        '\x1b[C': 'ARROW_RT',
        '\x1b[D': 'ARROW_LT',
    }
    return key if key not in keylist else keylist[key]


get_keypress.__module__ = 'rads'


def _prompt_timeout(signum, _frame):
    raise TimeoutError("Prompt timed out")


def prompt_y_n(
    question: str,
    *,
    timeout: Union[int, None] = 604800,
    sigint_exit: bool = True,
) -> bool:
    """Prompt for a yes or no answer

    Args:
        question: text to display before [y/n] prompt
        timeout: time in seconds after which to raise TimeoutError. Defaults
          to one week
        sigint_exit: Whether to sys.exit(1) on ctrl+c; Defaults True.

    Raises:
        SystemExit: if SIGINT is received
        KeyboardInterrupt: if sigint_exit=False and a sigint is received
        TimeoutError: if timeout= is set and no response in that time
    """
    valid = {"yes": True, "y": True, "no": False, "n": False}
    while True:
        print(question, '[y/n]')
        try:
            if timeout:
                signal.signal(signal.SIGALRM, _prompt_timeout)
                signal.alarm(timeout)
            choice = input().lower()
            if timeout:
                signal.alarm(0)  # disable the alarm
        except KeyboardInterrupt:
            if not sigint_exit:
                raise
            print("\nCancelled")
            sys.exit(1)
        if choice in valid:
            return valid[choice]
        print('Invalid answer. Try again:', color.bold('[y/n]'))


prompt_y_n.__module__ = 'rads'



"""
    :codeauthor: :email:`Alexandru Bleotu (alexandru.bleotu@morganstanley.com)`


    salt.config.schemas.esxdatacenter
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ESX Datacenter configuration schemas
"""

from salt.utils.schema import ArrayItem, IntegerItem, Schema, StringItem


class EsxdatacenterProxySchema(Schema):
    """
    Schema of the esxdatacenter proxy input
    """

    title = "Esxdatacenter Proxy Schema"
    description = "Esxdatacenter proxy schema"
    additional_properties = False
    proxytype = StringItem(required=True, enum=["esxdatacenter"])
    vcenter = StringItem(required=True, pattern=r"[^\s]+")
    datacenter = StringItem(required=True)
    mechanism = StringItem(required=True, enum=["userpass", "sspi"])
    username = StringItem()
    passwords = ArrayItem(min_items=1, items=StringItem(), unique_items=True)
    # TODO Should be changed when anyOf is supported for schemas
    domain = StringItem()
    principal = StringItem()
    protocol = StringItem()
    port = IntegerItem(minimum=1)



"""
Manage BTRFS file systems.

:maintainer:    Alberto Planas <aplanas@suse.com>
:maturity:      new
:depends:       None
:platform:      Linux
"""

import functools
import logging
import os.path
import tempfile

from salt.exceptions import CommandExecutionError

log = logging.getLogger(__name__)

__virtualname__ = "btrfs"


def _mount(device, use_default):
    """
    Mount the device in a temporary place.
    """
    opts = "defaults" if use_default else "subvol=/"
    dest = tempfile.mkdtemp()
    res = __states__["mount.mounted"](
        dest, device=device, fstype="btrfs", opts=opts, persist=False
    )
    if not res["result"]:
        log.error("Cannot mount device %s in %s", device, dest)
        _umount(dest)
        return None
    return dest


def _umount(path):
    """
    Umount and clean the temporary place.
    """
    __states__["mount.unmounted"](path)
    __utils__["files.rm_rf"](path)


def _is_default(path, dest, name):
    """
    Check if the subvolume is the current default.
    """
    subvol_id = __salt__["btrfs.subvolume_show"](path)[name]["subvolume id"]
    def_id = __salt__["btrfs.subvolume_get_default"](dest)["id"]
    return subvol_id == def_id


def _set_default(path, dest, name):
    """
    Set the subvolume as the current default.
    """
    subvol_id = __salt__["btrfs.subvolume_show"](path)[name]["subvolume id"]
    return __salt__["btrfs.subvolume_set_default"](subvol_id, dest)


def _is_cow(path):
    """
    Check if the subvolume is copy on write
    """
    dirname = os.path.dirname(path)
    return "C" not in __salt__["file.lsattr"](dirname)[path]


def _unset_cow(path):
    """
    Disable the copy on write in a subvolume
    """
    return __salt__["file.chattr"](path, operator="add", attributes="C")


def __mount_device(action):
    """
    Small decorator to makes sure that the mount and umount happends in
    a transactional way.
    """

    @functools.wraps(action)
    def wrapper(*args, **kwargs):
        name = kwargs.get("name", args[0] if args else None)
        device = kwargs.get("device", args[1] if len(args) > 1 else None)
        use_default = kwargs.get("use_default", False)

        ret = {
            "name": name,
            "result": False,
            "changes": {},
            "comment": ["Some error happends during the operation."],
        }
        try:
            if device:
                dest = _mount(device, use_default)
                if not dest:
                    msg = f"Device {device} cannot be mounted"
                    ret["comment"].append(msg)
                kwargs["__dest"] = dest
            ret = action(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            log.error("Exception raised while mounting device: %s", exc, exc_info=True)
            ret["comment"].append(exc)
        finally:
            if device:
                _umount(dest)
        return ret

    return wrapper


@__mount_device
def subvolume_created(
    name,
    device,
    qgroupids=None,
    set_default=False,
    copy_on_write=True,
    force_set_default=True,
    __dest=None,
):
    """
    Makes sure that a btrfs subvolume is present.

    name
        Name of the subvolume to add

    device
        Device where to create the subvolume

    qgroupids
         Add the newly created subcolume to a qgroup. This parameter
         is a list

    set_default
        If True, this new subvolume will be set as default when
        mounted, unless subvol option in mount is used

    copy_on_write
        If false, set the subvolume with chattr +C

    force_set_default
        If false and the subvolume is already present, it will not
        force it as default if ``set_default`` is True

    """
    ret = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": [],
    }
    path = os.path.join(__dest, name)

    exists = __salt__["btrfs.subvolume_exists"](path)
    if exists:
        ret["comment"].append(f"Subvolume {name} already present")

    # Resolve first the test case. The check is not complete, but at
    # least we will report if a subvolume needs to be created. Can
    # happend that the subvolume is there, but we also need to set it
    # as default, or persist in fstab.
    if __opts__["test"]:
        ret["result"] = None
        if not exists:
            ret["changes"][name] = f"Subvolume {name} will be created"
        return ret

    if not exists:
        # Create the directories where the subvolume lives
        _path = os.path.dirname(path)
        res = __states__["file.directory"](_path, makedirs=True)
        if not res["result"]:
            ret["comment"].append(f"Error creating {_path} directory")
            return ret

        try:
            __salt__["btrfs.subvolume_create"](name, dest=__dest, qgroupids=qgroupids)
        except CommandExecutionError:
            ret["comment"].append(f"Error creating subvolume {name}")
            return ret

        ret["changes"][name] = f"Created subvolume {name}"

    # If the volume was already present, we can opt-out the check for
    # default subvolume.
    if (
        (not exists or (exists and force_set_default))
        and set_default
        and not _is_default(path, __dest, name)
    ):
        ret["changes"][name + "_default"] = _set_default(path, __dest, name)

    if not copy_on_write and _is_cow(path):
        ret["changes"][name + "_no_cow"] = _unset_cow(path)

    ret["result"] = True
    return ret


@__mount_device
def subvolume_deleted(name, device, commit=False, __dest=None):
    """
    Makes sure that a btrfs subvolume is removed.

    name
        Name of the subvolume to remove

    device
        Device where to remove the subvolume

    commit
        Wait until the transaction is over

    """
    ret = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": [],
    }

    path = os.path.join(__dest, name)

    exists = __salt__["btrfs.subvolume_exists"](path)
    if not exists:
        ret["comment"].append(f"Subvolume {name} already missing")

    if __opts__["test"]:
        ret["result"] = None
        if exists:
            ret["changes"][name] = f"Subvolume {name} will be removed"
        return ret

    # If commit is set, we wait until all is over
    commit = "after" if commit else None

    if not exists:
        try:
            __salt__["btrfs.subvolume_delete"](path, commit=commit)
        except CommandExecutionError:
            ret["comment"].append(f"Error removing subvolume {name}")
            return ret

        ret["changes"][name] = f"Removed subvolume {name}"

    ret["result"] = True
    return ret


def _diff_properties(expected, current):
    """Calculate the difference between the current and the expected
    properties

    * 'expected' is expressed in a dictionary like: {'property': value}

    * 'current' contains the same format retuned by 'btrfs.properties'

    If the property is not available, will throw an exception.

    """
    difference = {}
    for _property, value in expected.items():
        current_value = current[_property]["value"]
        if value is False and current_value == "N/A":
            needs_update = False
        elif value != current_value:
            needs_update = True
        else:
            needs_update = False
        if needs_update:
            difference[_property] = value
    return difference


@__mount_device
def properties(name, device, use_default=False, __dest=None, **properties):
    """
    Makes sure that a list of properties are set in a subvolume, file
    or device.

    name
        Name of the object to change

    device
        Device where the object lives, if None, the device will be in
        name

    use_default
        If True, this subvolume will be resolved to the default
        subvolume assigned during the create operation

    properties
        Dictionary of properties

    Valid properties are 'ro', 'label' or 'compression'. Check the
    documentation to see where those properties are valid for each
    object.

    """
    ret = {
        "name": name,
        "result": False,
        "changes": {},
        "comment": [],
    }

    # 'name' will have always the name of the object that we want to
    # change, but if the object is a device, we do not repeat it again
    # in 'device'. This makes device sometimes optional.
    if device:
        if os.path.isabs(name):
            path = os.path.join(__dest, os.path.relpath(name, os.path.sep))
        else:
            path = os.path.join(__dest, name)
    else:
        path = name

    if not os.path.exists(path):
        ret["comment"].append(f"Object {name} not found")
        return ret

    # Convert the booleans to lowercase
    properties = {
        k: v if type(v) is not bool else str(v).lower() for k, v in properties.items()
    }

    current_properties = {}
    try:
        current_properties = __salt__["btrfs.properties"](path)
    except CommandExecutionError as e:
        ret["comment"].append(f"Error reading properties from {name}")
        ret["comment"].append(f"Current error {e}")
        return ret

    try:
        properties_to_set = _diff_properties(properties, current_properties)
    except KeyError:
        ret["comment"].append(f"Some property not found in {name}")
        return ret

    if __opts__["test"]:
        ret["result"] = None
        if properties_to_set:
            ret["changes"] = properties_to_set
        else:
            msg = f"No properties will be changed in {name}"
            ret["comment"].append(msg)
        return ret

    if properties_to_set:
        _properties = ",".join(f"{k}={v}" for k, v in properties_to_set.items())
        __salt__["btrfs.properties"](path, set=_properties)

        current_properties = __salt__["btrfs.properties"](path)
        properties_failed = _diff_properties(properties, current_properties)
        if properties_failed:
            msg = "Properties {} failed to be changed in {}".format(
                properties_failed, name
            )
            ret["comment"].append(msg)
            return ret

        ret["comment"].append(f"Properties changed in {name}")
        ret["changes"] = properties_to_set
    else:
        ret["comment"].append(f"Properties not changed in {name}")

    ret["result"] = True
    return ret



"""
Interact with virtual machine images via libguestfs

:depends:   - libguestfs
"""

import hashlib
import logging
import os
import tempfile
import time

import salt.utils.path
from salt.config import DEFAULT_HASH_TYPE

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if libguestfs python bindings are installed
    """
    if salt.utils.path.which("guestmount"):
        return "guestfs"
    return (
        False,
        "The guestfs execution module cannot be loaded: guestmount binary not in path.",
    )


def mount(location, access="rw", root=None):
    """
    Mount an image

    CLI Example:

    .. code-block:: bash

        salt '*' guest.mount /srv/images/fedora.qcow
    """
    if root is None:
        root = os.path.join(
            tempfile.gettempdir(), "guest", location.lstrip(os.sep).replace("/", ".")
        )
        log.debug("Using root %s", root)
    if not os.path.isdir(root):
        try:
            os.makedirs(root)
        except OSError:
            # Somehow the path already exists
            pass
    while True:
        if os.listdir(root):
            # Stuff is in there, don't use it
            hash_type = getattr(hashlib, __opts__.get("hash_type", DEFAULT_HASH_TYPE))
            rand = hash_type(os.urandom(32)).hexdigest()
            root = os.path.join(
                tempfile.gettempdir(),
                "guest",
                location.lstrip(os.sep).replace("/", ".") + rand,
            )
            log.debug("Establishing new root as %s", root)
            if not os.path.isdir(root):
                try:
                    os.makedirs(root)
                except OSError:
                    log.info("Path already existing: %s", root)
        else:
            break
    cmd = f"guestmount -i -a {location} --{access} {root}"
    __salt__["cmd.run"](cmd, python_shell=False)
    return root


def umount(name, disk=None):
    """
    Unmount an image

    CLI Example:

    .. code-block:: bash

        salt '*' guestfs.umount /mountpoint disk=/srv/images/fedora.qcow
    """
    cmd = f"guestunmount -q {name}"
    __salt__["cmd.run"](cmd)

    # Wait at most 5s that the disk is no longuer used
    loops = 0
    while (
        disk is not None
        and loops < 5
        and len(__salt__["cmd.run"](f"lsof {disk}").splitlines()) != 0
    ):
        loops = loops + 1
        time.sleep(1)



"""
Module for freezer
:maintainer:    Alberto Planas <aplanas@suse.com>
:maturity:      new
:depends:       None
:platform:      Linux
"""

import logging
import os

import salt.utils.dictdiffer
import salt.utils.json as json
from salt.exceptions import CommandExecutionError
from salt.utils.args import clean_kwargs
from salt.utils.files import fopen

log = logging.getLogger(__name__)

__func_alias__ = {
    "list_": "list",
}


def __virtual__():
    """
    Freezer is based on top of the pkg module.

    Return True since the `pkg` module should always exist. This
    avoids the overhead of loading all modules.

    """
    return True


def _states_path():
    """
    Return the path where we will store the states.
    """
    return os.path.join(__opts__["cachedir"], "freezer")


def _paths(name=None):
    """
    Return the full path for the packages and repository freezer
    files.

    """
    name = "freezer" if not name else name
    states_path = _states_path()
    return (
        os.path.join(states_path, f"{name}-pkgs.yml"),
        os.path.join(states_path, f"{name}-reps.yml"),
    )


def status(name=None):
    """
    Return True if there is already a frozen state.

    A frozen state is merely a list of packages (including the
    version) in a specific time. This information can be used to
    compare with the current list of packages, and revert the
    installation of some extra packages that are in the system.

    name
        Name of the frozen state. Optional.

    CLI Example:

    .. code-block:: bash

        salt '*' freezer.status
        salt '*' freezer.status pre_install

    """
    name = "freezer" if not name else name
    return all(os.path.isfile(i) for i in _paths(name))


def list_():
    """
    Return the list of frozen states.

    CLI Example:

    .. code-block:: bash

        salt '*' freezer.list

    """
    ret = []
    states_path = _states_path()
    if not os.path.isdir(states_path):
        return ret

    for state in os.listdir(states_path):
        if state.endswith(("-pkgs.yml", "-reps.yml")):
            # Remove the suffix - both files start with the freezer
            # name
            ret.append(state[:-9])
    return sorted(set(ret))


def freeze(name=None, force=False, **kwargs):
    """
    Save the list of package and repos in a freeze file.

    As this module is build on top of the pkg module, the user can
    send extra attributes to the underlying pkg module via kwargs.
    This function will call ``pkg.list_pkgs`` and ``pkg.list_repos``,
    and any additional arguments will be passed through to those
    functions.

    name
        Name of the frozen state. Optional.

    force
        If true, overwrite the state. Optional.

    CLI Example:

    .. code-block:: bash

        salt '*' freezer.freeze
        salt '*' freezer.freeze pre_install
        salt '*' freezer.freeze force=True root=/chroot

    """
    states_path = _states_path()

    try:
        if not os.path.exists(states_path):
            os.makedirs(states_path)
    except OSError as e:
        msg = "Error when trying to create the freezer storage %s: %s"
        log.error(msg, states_path, e)
        raise CommandExecutionError(msg % (states_path, e))

    if status(name) and not force:
        raise CommandExecutionError(
            "The state is already present. Use force parameter to overwrite."
        )
    safe_kwargs = clean_kwargs(**kwargs)
    pkgs = __salt__["pkg.list_pkgs"](**safe_kwargs)
    repos = __salt__["pkg.list_repos"](**safe_kwargs)
    for fname, content in zip(_paths(name), (pkgs, repos)):
        with fopen(fname, "w") as fp:
            json.dump(content, fp)
    return True


def _add_missing_repositories(frozen_repos, ret, **kwargs):
    """Add missing repositories and update the ret dict"""
    repos = __salt__["pkg.list_repos"](**kwargs)
    missing_repos = set(frozen_repos) - set(repos)
    for repo in missing_repos:
        try:
            # In Python 2 we cannot do advance destructuring, so we
            # need to create a temporary dictionary that will merge
            # all the parameters
            _tmp_kwargs = frozen_repos[repo].copy()
            _tmp_kwargs.update(kwargs)
            __salt__["pkg.mod_repo"](repo, **_tmp_kwargs)
            ret["repos"]["add"].append(repo)
            log.info("Added missing repository %s", repo)
        except Exception as e:  # pylint: disable=broad-except
            msg = "Error adding %s repository: %s"
            log.error(msg, repo, e)
            ret["comment"].append(msg % (repo, e))


def _remove_extra_repositories(frozen_repos, ret, **kwargs):
    """Remove extra repositories and update the ret dict"""
    repos = __salt__["pkg.list_repos"](**kwargs)
    extra_repos = set(repos) - set(frozen_repos)
    for repo in extra_repos:
        try:
            __salt__["pkg.del_repo"](repo, **kwargs)
            ret["repos"]["remove"].append(repo)
            log.info("Removed extra repository %s", repo)
        except Exception as e:  # pylint: disable=broad-except
            msg = "Error removing %s repository: %s"
            log.error(msg, repo, e)
            ret["comment"].append(msg % (repo, e))


def _add_missing_packages(frozen_pkgs, ret, **kwargs):
    """Add missing packages and update the ret dict"""
    # NOTE: we can remove the `for` using `pkgs`. This will improve
    # performance, but I want to have a more detalied report of what
    # packages are installed or failed.
    pkgs = __salt__["pkg.list_pkgs"](**kwargs)
    missing_pkgs = set(frozen_pkgs) - set(pkgs)
    for pkg in missing_pkgs:
        try:
            __salt__["pkg.install"](name=pkg, **kwargs)
            ret["pkgs"]["add"].append(pkg)
            log.info("Added missing package %s", pkg)
        except Exception as e:  # pylint: disable=broad-except
            msg = "Error adding %s package: %s"
            log.error(msg, pkg, e)
            ret["comment"].append(msg % (pkg, e))


def _remove_extra_packages(frozen_pkgs, ret, **kwargs):
    """Remove extra packages and update the ret dict"""
    pkgs = __salt__["pkg.list_pkgs"](**kwargs)
    extra_pkgs = set(pkgs) - set(frozen_pkgs)
    for pkg in extra_pkgs:
        try:
            __salt__["pkg.remove"](name=pkg, **kwargs)
            ret["pkgs"]["remove"].append(pkg)
            log.info("Removed extra package %s", pkg)
        except Exception as e:  # pylint: disable=broad-except
            msg = "Error removing %s package: %s"
            log.error(msg, pkg, e)
            ret["comment"].append(msg % (pkg, e))


def restore(name=None, clean=False, **kwargs):
    """
    Make sure that the system contains the packages and repos from a
    frozen state.

    Read the list of packages and repositories from the freeze file,
    and compare it with the current list of packages and repos. If
    there is any difference, all the missing packages are repos will
    be installed, and all the extra packages and repos will be
    removed.

    As this module is build on top of the pkg module, the user can
    send extra attributes to the underlying pkg module via kwargs.
    This function will call ``pkg.list_repos``, ``pkg.mod_repo``,
    ``pkg.list_pkgs``, ``pkg.install``, ``pkg.remove`` and
    ``pkg.del_repo``, and any additional arguments will be passed
    through to those functions.

    name
        Name of the frozen state. Optional.

    clean
        If True remove the frozen information YAML from the cache

        .. versionadded:: 3000

    CLI Example:

    .. code-block:: bash

        salt '*' freezer.restore
        salt '*' freezer.restore root=/chroot

    """
    if not status(name):
        raise CommandExecutionError("Frozen state not found.")

    frozen_pkgs = {}
    frozen_repos = {}
    for fname, content in zip(_paths(name), (frozen_pkgs, frozen_repos)):
        with fopen(fname) as fp:
            content.update(json.load(fp))

    # The ordering of removing or adding packages and repos can be
    # relevant, as maybe some missing package comes from a repo that
    # is also missing, so it cannot be installed. But can also happend
    # that a missing package comes from a repo that is present, but
    # will be removed.
    #
    # So the proposed order is;
    #   - Add missing repos
    #   - Add missing packages
    #   - Remove extra packages
    #   - Remove extra repos

    safe_kwargs = clean_kwargs(**kwargs)

    # Note that we expect that the information stored in list_XXX
    # match with the mod_XXX counterpart. If this is not the case the
    # recovery will be partial.

    ret = {
        "pkgs": {"add": [], "remove": []},
        "repos": {"add": [], "remove": []},
        "comment": [],
    }

    _add_missing_repositories(frozen_repos, ret, **safe_kwargs)
    _add_missing_packages(frozen_pkgs, ret, **safe_kwargs)
    _remove_extra_packages(frozen_pkgs, ret, **safe_kwargs)
    _remove_extra_repositories(frozen_repos, ret, **safe_kwargs)

    # Clean the cached YAML files
    if clean and not ret["comment"]:
        for fname in _paths(name):
            os.remove(fname)

    return ret


def compare(old, new):
    """
    Display the difference between two frozen states. The results are shown as
    as a dictionary with keys for packages and repositories. Each key may
    contain a changes dictionary showing items that differ between the two
    frozen states. Items shown in the "old" changes but not the "new" were
    removed. Items in "new" but not "old" were added. Items shown in both
    probably updated/changed versions between freezes.

    old
        Name of the "old" frozen state. Required.

    new
        Name of the "new" frozen state. Required.

    CLI Example:

    .. code-block:: bash

        salt '*' freezer.freeze pre_install post_install

    """
    ret = {}

    if not (status(old) and status(new)):
        raise CommandExecutionError("Frozen state not found.")

    for ofile, nfile in zip(_paths(old), _paths(new)):
        with fopen(ofile, "r") as ofp:
            old_dict = json.load(ofp)
        with fopen(nfile, "r") as nfp:
            new_dict = json.load(nfp)
        if ofile.endswith("-pkgs.yml"):
            ret["pkgs"] = salt.utils.dictdiffer.deep_diff(old_dict, new_dict)
        elif ofile.endswith("-reps.yml"):
            ret["repos"] = salt.utils.dictdiffer.deep_diff(old_dict, new_dict)

    return ret



"""
Install software from the FreeBSD ``ports(7)`` system

.. versionadded:: 2014.1.0

This module allows you to install ports using ``BATCH=yes`` to bypass
configuration prompts. It is recommended to use the :mod:`ports state
<salt.states.freebsdports>` to install ports, but it is also possible to use
this module exclusively from the command line.

.. code-block:: bash

    salt minion-id ports.config security/nmap IPV6=off
    salt minion-id ports.install security/nmap
"""

import fnmatch
import logging
import os
import re

import salt.utils.data
import salt.utils.files
import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "ports"


def __virtual__():
    """
    Only runs on FreeBSD systems
    """
    if __grains__["os"] == "FreeBSD":
        return __virtualname__
    return (
        False,
        "The freebsdports execution module cannot be loaded: "
        "only available on FreeBSD systems.",
    )


def _portsnap():
    """
    Return 'portsnap --interactive' for FreeBSD 10, otherwise 'portsnap'
    """
    ret = ["portsnap"]
    if float(__grains__["osrelease"]) >= 10:
        ret.append("--interactive")
    return ret


def _check_portname(name):
    """
    Check if portname is valid and whether or not the directory exists in the
    ports tree.
    """
    if not isinstance(name, str) or "/" not in name:
        raise SaltInvocationError(f"Invalid port name '{name}' (category required)")

    path = os.path.join("/usr/ports", name)
    if not os.path.isdir(path):
        raise SaltInvocationError(f"Path '{path}' does not exist")

    return path


def _options_dir(name):
    """
    Retrieve the path to the dir containing OPTIONS file for a given port
    """
    _check_portname(name)
    _root = "/var/db/ports"

    # New path: /var/db/ports/category_portname
    new_dir = os.path.join(_root, name.replace("/", "_"))
    # Old path: /var/db/ports/portname
    old_dir = os.path.join(_root, name.split("/")[-1])

    if os.path.isdir(old_dir):
        return old_dir
    return new_dir


def _options_file_exists(name):
    """
    Returns True/False based on whether or not the options file for the
    specified port exists.
    """
    return os.path.isfile(os.path.join(_options_dir(name), "options"))


def _write_options(name, configuration):
    """
    Writes a new OPTIONS file
    """
    _check_portname(name)

    pkg = next(iter(configuration))
    conf_ptr = configuration[pkg]

    dirname = _options_dir(name)
    if not os.path.isdir(dirname):
        try:
            os.makedirs(dirname)
        except OSError as exc:
            raise CommandExecutionError(f"Unable to make {dirname}: {exc}")

    with salt.utils.files.fopen(os.path.join(dirname, "options"), "w") as fp_:
        sorted_options = list(conf_ptr)
        sorted_options.sort()
        fp_.write(
            salt.utils.stringutils.to_str(
                "# This file was auto-generated by Salt (http://saltstack.com)\n"
                "# Options for {0}\n"
                "_OPTIONS_READ={0}\n"
                "_FILE_COMPLETE_OPTIONS_LIST={1}\n".format(
                    pkg, " ".join(sorted_options)
                )
            )
        )
        opt_tmpl = "OPTIONS_FILE_{0}SET+={1}\n"
        for opt in sorted_options:
            fp_.write(
                salt.utils.stringutils.to_str(
                    opt_tmpl.format("" if conf_ptr[opt] == "on" else "UN", opt)
                )
            )


def _normalize(val):
    """
    Fix Salt's yaml-ification of on/off, and otherwise normalize the on/off
    values to be used in writing the options file
    """
    if isinstance(val, bool):
        return "on" if val else "off"
    return str(val).lower()


def install(name, clean=True):
    """
    Install a port from the ports tree. Installs using ``BATCH=yes`` for
    non-interactive building. To set config options for a given port, use
    :mod:`ports.config <salt.modules.freebsdports.config>`.

    clean : True
        If ``True``, cleans after installation. Equivalent to running ``make
        install clean BATCH=yes``.

    .. note::

        It may be helpful to run this function using the ``-t`` option to set a
        higher timeout, since compiling a port may cause the Salt command to
        exceed the default timeout.

    CLI Example:

    .. code-block:: bash

        salt -t 1200 '*' ports.install security/nmap
    """
    portpath = _check_portname(name)
    old = __salt__["pkg.list_pkgs"]()
    if old.get(name.rsplit("/")[-1]):
        deinstall(name)
    cmd = ["make", "install"]
    if clean:
        cmd.append("clean")
    cmd.append("BATCH=yes")
    result = __salt__["cmd.run_all"](
        cmd, cwd=portpath, reset_system_locale=False, python_shell=False
    )
    if result["retcode"] != 0:
        __context__["ports.install_error"] = result["stderr"]
    __context__.pop("pkg.list_pkgs", None)
    new = __salt__["pkg.list_pkgs"]()
    ret = salt.utils.data.compare_dicts(old, new)
    if not ret and result["retcode"] == 0:
        # No change in package list, but the make install was successful.
        # Assume that the installation was a recompile with new options, and
        # set return dict so that changes are detected by the ports.installed
        # state.
        ret = {name: {"old": old.get(name, ""), "new": new.get(name, "")}}
    return ret


def deinstall(name):
    """
    De-install a port.

    CLI Example:

    .. code-block:: bash

        salt '*' ports.deinstall security/nmap
    """
    portpath = _check_portname(name)
    old = __salt__["pkg.list_pkgs"]()
    result = __salt__["cmd.run_all"](
        ["make", "deinstall", "BATCH=yes"], cwd=portpath, python_shell=False
    )
    __context__.pop("pkg.list_pkgs", None)
    new = __salt__["pkg.list_pkgs"]()
    return salt.utils.data.compare_dicts(old, new)


def rmconfig(name):
    """
    Clear the cached options for the specified port; run a ``make rmconfig``

    name
        The name of the port to clear

    CLI Example:

    .. code-block:: bash

        salt '*' ports.rmconfig security/nmap
    """
    portpath = _check_portname(name)
    return __salt__["cmd.run"](["make", "rmconfig"], cwd=portpath, python_shell=False)


def showconfig(name, default=False, dict_return=False):
    """
    Show the configuration options for a given port.

    default : False
        Show the default options for a port (not necessarily the same as the
        current configuration)

    dict_return : False
        Instead of returning the output of ``make showconfig``, return the data
        in an dictionary

    CLI Example:

    .. code-block:: bash

        salt '*' ports.showconfig security/nmap
        salt '*' ports.showconfig security/nmap default=True
    """
    portpath = _check_portname(name)

    if default and _options_file_exists(name):
        saved_config = showconfig(name, default=False, dict_return=True)
        rmconfig(name)
        if _options_file_exists(name):
            raise CommandExecutionError("Unable to get default configuration")
        default_config = showconfig(name, default=False, dict_return=dict_return)
        _write_options(name, saved_config)
        return default_config

    try:
        result = __salt__["cmd.run_all"](
            ["make", "showconfig"], cwd=portpath, python_shell=False
        )
        output = result["stdout"].splitlines()
        if result["retcode"] != 0:
            error = result["stderr"]
        else:
            error = ""
    except TypeError:
        error = result

    if error:
        msg = f"Error running 'make showconfig' for {name}: {error}"
        log.error(msg)
        raise SaltInvocationError(msg)

    if not dict_return:
        return "\n".join(output)

    if (not output) or ("configuration options" not in output[0]):
        return {}

    try:
        pkg = output[0].split()[-1].rstrip(":")
    except (IndexError, AttributeError, TypeError) as exc:
        log.error("Unable to get pkg-version string: %s", exc)
        return {}

    ret = {pkg: {}}
    output = output[1:]
    for line in output:
        try:
            opt, val, desc = re.match(r"\s+([^=]+)=(off|on): (.+)", line).groups()
        except AttributeError:
            continue
        ret[pkg][opt] = val

    if not ret[pkg]:
        return {}
    return ret


def config(name, reset=False, **kwargs):
    """
    Modify configuration options for a given port. Multiple options can be
    specified. To see the available options for a port, use
    :mod:`ports.showconfig <salt.modules.freebsdports.showconfig>`.

    name
        The port name, in ``category/name`` format

    reset : False
        If ``True``, runs a ``make rmconfig`` for the port, clearing its
        configuration before setting the desired options

    CLI Examples:

    .. code-block:: bash

        salt '*' ports.config security/nmap IPV6=off
    """
    portpath = _check_portname(name)

    if reset:
        rmconfig(name)

    configuration = showconfig(name, dict_return=True)

    if not configuration:
        raise CommandExecutionError(f"Unable to get port configuration for '{name}'")

    # Get top-level key for later reference
    pkg = next(iter(configuration))
    conf_ptr = configuration[pkg]

    opts = {str(x): _normalize(kwargs[x]) for x in kwargs if not x.startswith("_")}

    bad_opts = [x for x in opts if x not in conf_ptr]
    if bad_opts:
        raise SaltInvocationError(
            "The following opts are not valid for port {}: {}".format(
                name, ", ".join(bad_opts)
            )
        )

    bad_vals = [f"{x}={y}" for x, y in opts.items() if y not in ("on", "off")]
    if bad_vals:
        raise SaltInvocationError(
            "The following key/value pairs are invalid: {}".format(", ".join(bad_vals))
        )

    conf_ptr.update(opts)
    _write_options(name, configuration)

    new_config = showconfig(name, dict_return=True)
    try:
        new_config = new_config[next(iter(new_config))]
    except (StopIteration, TypeError):
        return False

    return all(conf_ptr[x] == new_config.get(x) for x in conf_ptr)


def update(extract=False):
    """
    Update the ports tree

    extract : False
        If ``True``, runs a ``portsnap extract`` after fetching, should be used
        for first-time installation of the ports tree.

    CLI Example:

    .. code-block:: bash

        salt '*' ports.update
    """
    result = __salt__["cmd.run_all"](_portsnap() + ["fetch"], python_shell=False)
    if not result["retcode"] == 0:
        raise CommandExecutionError(
            "Unable to fetch ports snapshot: {}".format(result["stderr"])
        )

    ret = []
    try:
        patch_count = re.search(r"Fetching (\d+) patches", result["stdout"]).group(1)
    except AttributeError:
        patch_count = 0

    try:
        new_port_count = re.search(
            r"Fetching (\d+) new ports or files", result["stdout"]
        ).group(1)
    except AttributeError:
        new_port_count = 0

    ret.append(f"Applied {patch_count} new patches")
    ret.append(f"Fetched {new_port_count} new ports or files")

    if extract:
        result = __salt__["cmd.run_all"](_portsnap() + ["extract"], python_shell=False)
        if not result["retcode"] == 0:
            raise CommandExecutionError(
                "Unable to extract ports snapshot {}".format(result["stderr"])
            )

    result = __salt__["cmd.run_all"](_portsnap() + ["update"], python_shell=False)
    if not result["retcode"] == 0:
        raise CommandExecutionError(
            "Unable to apply ports snapshot: {}".format(result["stderr"])
        )

    __context__.pop("ports.list_all", None)
    return "\n".join(ret)


def list_all():
    """
    Lists all ports available.

    CLI Example:

    .. code-block:: bash

        salt '*' ports.list_all

    .. warning::

        Takes a while to run, and returns a **LOT** of output
    """
    if "ports.list_all" not in __context__:
        __context__["ports.list_all"] = []
        for path, dirs, files in salt.utils.path.os_walk("/usr/ports"):
            stripped = path[len("/usr/ports") :]
            if stripped.count("/") != 2 or stripped.endswith("/CVS"):
                continue
            __context__["ports.list_all"].append(stripped[1:])
    return __context__["ports.list_all"]


def search(name):
    """
    Search for matches in the ports tree. Globs are supported, and the category
    is optional

    CLI Examples:

    .. code-block:: bash

        salt '*' ports.search 'security/*'
        salt '*' ports.search 'security/n*'
        salt '*' ports.search nmap

    .. warning::

        Takes a while to run
    """
    name = str(name)
    all_ports = list_all()
    if "/" in name:
        if name.count("/") > 1:
            raise SaltInvocationError(
                "Invalid search string '{0}'. Port names cannot have more "
                "than one slash"
            )
        else:
            return fnmatch.filter(all_ports, name)
    else:
        ret = []
        for port in all_ports:
            if fnmatch.fnmatch(port.rsplit("/")[-1], name):
                ret.append(port)
        return ret
